==> reboot_resemble-0.5.3-cp310-cp310-macosx_13_0_x86_64.whl: resemble/integrations/mailgun/servicers.py <==
import aiohttp
import asyncio
import os
import resemble.aio.auth.token_verifiers
import respect.logging
import uuid
from abc import abstractmethod
from datetime import datetime, timedelta
from google.protobuf.timestamp_pb2 import Timestamp
from resemble.aio.aborted import SystemAborted
from resemble.aio.auth import Auth
from resemble.aio.auth.tokens import Token
from resemble.aio.contexts import ReaderContext, WriterContext
from resemble.aio.secrets import SecretNotFoundException, Secrets
from resemble.aio.tasks import TaskEffect
from resemble.integrations.mailgun import (
    MAILGUN_EVENT_API_CONSISTENCY_DELAY_SEC,
)
from resemble.integrations.mailgun.v1.mailgun_rsm import (
    Message,
    MessageState,
    SendRequest,
    SendResponse,
    SendTaskRequest,
    SendTaskResponse,
)
from resemble.v1alpha1.errors_pb2 import Unauthenticated
from typing import ClassVar, Optional

logger = respect.logging.get_logger(__name__)

IDEMPOTENCY_KEY_USER_VARIABLE = "idempotency_message_key"
MAILGUN_API_KEY_SECRET_NAME = "mailgun-api-key"
MAILGUN_API_BASE_ADDRESS_ENV_NAME = "MAILGUN_API_BASE_ADDRESS"

# If no API base address is injected for testing, use this one (the address for
# the real Mailgun API).
MAILGUN_API_BASE_ADDRESS = "https://api.mailgun.net"

# A "tag" used to ensure that the mailgun API key was correctly
# verified.
MAILGUN_API_KEY_VERIFIED = 'MAILGUN_API_KEY_VERIFIED'


class MailgunAPIError(RuntimeError):
    pass


class _AbstractMessageServicer(Message.Interface):

    class TokenVerifier(resemble.aio.auth.token_verifiers.TokenVerifier):

        def __init__(self, secrets: Secrets):
            self._secrets = secrets

        async def verify_token(
            self,
            context: ReaderContext,
            token: str,
        ) -> Optional[Auth]:
            try:
                api_key = await self._secrets.get(MAILGUN_API_KEY_SECRET_NAME)
            except SecretNotFoundException:
                logger.critical(
                    "Failed to verify token, missing secret "
                    f"'{MAILGUN_API_KEY_SECRET_NAME}'"
                )
                return None
            else:
                if token == api_key.decode():
                    return Auth(
                        # No user ID, the secret _is_ the mailgun API key.
                        user_id=None,
                        bearer_token=Token(
                            raw=token,
                            decoded={MAILGUN_API_KEY_VERIFIED: True},
                        ),
                    )
                else:
                    return None

    class Authorizer(Message.Authorizer):

        async def authorize(
            self,
            *,
            method_name: str,
            context: ReaderContext,
            state: Optional[Message.Authorizer.StateType] = None,
            request: Optional[Message.Authorizer.RequestTypes] = None,
        ) -> bool:
            if context.auth is None:
                raise SystemAborted(Unauthenticated())

            # Invariant: our `_AbstractMessageServicer.TokenVerifier`
            # is verfiying that we have the correct API key, which is
            # sufficient for authorization, so all we need to do here
            # is ensure that we don't introduce a bug where a
            # different token verifier gets used. We do that by
            # looking for a specific field set by
            # `_AbstractMessageServicer.TokenVerifier`.
            assert context.auth.bearer_token is not None
            assert context.auth.bearer_token.decoded is not None
            return MAILGUN_API_KEY_VERIFIED in context.auth.bearer_token.decoded

    def __init__(self):
        self._secrets = Secrets()

    def token_verifier(
        self,
    ) -> Optional[resemble.aio.auth.token_verifiers.TokenVerifier]:
        return _AbstractMessageServicer.TokenVerifier(self._secrets)

    def authorizer(self) -> Optional[Message.Authorizer]:
        return _AbstractMessageServicer.Authorizer()

    async def Send(
        self,
        context: WriterContext,
        request: SendRequest,
    ) -> Message.SendEffects:
        initial_send_time = Timestamp()
        # GetCurrentTime() updates the Timestamp on which it is called in-place.
        initial_send_time.GetCurrentTime()

        # Generate an idempotency key to associate with the sending of this
        # message.
        state = MessageState(
            idempotency_key=str(uuid.uuid4()),
            initial_send_time=initial_send_time,
        )
        task: TaskEffect = self.schedule().SendTask(
            context,
            send_request=request,
        )
        return Message.SendEffects(
            state=state,
            tasks=[task],
            response=SendResponse(task_id=task.task_id),
        )

    async def SendTask(
        self,
        context: WriterContext,
        state: MessageState,
        request: SendTaskRequest,
    ) -> Message.SendTaskEffects:
        # Mailgun retains detailed event data for one day for free accounts and
        # up to 30 days for paid accounts based on subscription plan.
        # Since detailed event data is necessary for checking if an email has
        # already been sent, we cannot guarantee idempotency after more than a
        # day.
        #
        # Since we don't know if we're a free account or a paid
        # account, we conservatively bail out of this task if it's
        # been more than 24 hours since it first ran, to avoid
        # potentially sending this email twice.
        if (
            datetime.now()
            > state.initial_send_time.ToDatetime() + timedelta(hours=24)
        ):
            raise ValueError("Task is only valid for 24 hours after creation")

        text: Optional[str] = (
            request.send_request.text
            if request.send_request.HasField('text') else None
        )

        html: Optional[str] = (
            request.send_request.html
            if request.send_request.HasField('html') else None
        )

        if (
            (text is not None and html is not None) or
            (text is None and html is None)
        ):
            raise ValueError(
                "Exactly one of 'text' or 'html' should be specified"
            )

        body = text or html
        body_type = 'text' if text is not None else 'html'

        assert body is not None

        await self.send_email_idempotently(
            api_base_address=os.getenv(
                MAILGUN_API_BASE_ADDRESS_ENV_NAME, MAILGUN_API_BASE_ADDRESS
            ),
            sender=request.send_request.sender,
            recipient=request.send_request.recipient,
            subject=request.send_request.subject,
            domain=request.send_request.domain,
            idempotency_key=state.idempotency_key,
            body=body,
            body_type=body_type,
        )

        # Update accepted time to _now_, so when debugging we can see
        # when the message was accepted by mailgun (or if it has not
        # yet been accepted).
        #
        # `GetCurrentTime()` updates the `Timestamp` on which it is
        # called in-place.
        state.accepted_time.GetCurrentTime()

        return Message.SendTaskEffects(
            state=state,
            response=SendTaskResponse(),
        )

    @abstractmethod
    async def send_email(
        self,
        *,
        api_base_address: str,
        sender: str,
        recipient: str,
        subject: str,
        domain: str,
        idempotency_key: str,
        body: str,
        body_type: str,
    ):
        pass

    @abstractmethod
    async def is_email_with_idempotency_key_sent_to_recipient(
        self,
        *,
        api_base_address: str,
        recipient: str,
        domain: str,
        idempotency_key: str,
    ) -> bool:
        pass

    async def send_email_idempotently(
        self,
        *,
        api_base_address: str,
        sender: str,
        recipient: str,
        subject: str,
        domain: str,
        idempotency_key: str,
        body: str,
        body_type: str,
    ):
        # After our delay for eventual consistency (see comment on constant
        # declaration), we assume that the endpoint has fully caught up. If it
        # does not have evidence that the email has already been sent, we send
        # it now.
        await asyncio.sleep(MAILGUN_EVENT_API_CONSISTENCY_DELAY_SEC)

        if (
            not await self.is_email_with_idempotency_key_sent_to_recipient(
                api_base_address=api_base_address,
                recipient=recipient,
                domain=domain,
                idempotency_key=idempotency_key,
            )
        ):
            await self.send_email(
                api_base_address=api_base_address,
                sender=sender,
                recipient=recipient,
                subject=subject,
                domain=domain,
                body=body,
                body_type=body_type,
                idempotency_key=idempotency_key,
            )


class MessageServicer(_AbstractMessageServicer):

    def __init__(self):
        super().__init__()

    async def get_api_key(self) -> str:
        try:
            api_key = await self._secrets.get(MAILGUN_API_KEY_SECRET_NAME)
        except SecretNotFoundException as exception:
            logger.critical(f"No secret '{MAILGUN_API_KEY_SECRET_NAME}'")
            raise RuntimeError("No API key for mailgun") from exception
        return api_key.decode()

    async def send_email(
        self,
        *,
        api_base_address: str,
        sender: str,
        recipient: str,
        subject: str,
        domain: str,
        idempotency_key: str,
        body: str,
        body_type: str,
    ):
        """
        Sends an email with the given subject and body from the sender to
        the recipient, using the Mailgun API at the given api_base_address and
        domain. Authenticates to Mailgun with the given api_key credential.

        The email body can be provided either as plain text or as HTML content.
        We add the 'idempotency_key' as a user variable on the Mailgun message
        so that we can uniquely identify this email and ensure it is only sent
        once."""
        api_key = await self.get_api_key()

        async with aiohttp.ClientSession(
            auth=aiohttp.BasicAuth(login="api", password=api_key)
        ) as session:
            async with session.post(
                f"{api_base_address}/v3/{domain}/messages",
                data={
                    "from": sender,
                    "to": recipient,
                    "subject": subject,
                    body_type: body,
                    f"v:{IDEMPOTENCY_KEY_USER_VARIABLE}": idempotency_key,
                }
            ) as response:
                if response.status != 200:
                    raise MailgunAPIError(
                        f"Something went wrong while running POST on Mailgun API, error {response.status}"
                    )

                logger.info(
                    "Mailgun accepted email message for "
                    f"recipient '{recipient}' with idempotency key "
                    f"'{idempotency_key}'"
                )

    async def is_email_with_idempotency_key_sent_to_recipient(
        self,
        *,
        api_base_address: str,
        recipient: str,
        domain: str,
        idempotency_key: str,
    ) -> bool:
        api_key = await self.get_api_key()

        async with aiohttp.ClientSession(
            auth=aiohttp.BasicAuth(login="api", password=api_key)
        ) as session:
            user_variables = f"{{'{IDEMPOTENCY_KEY_USER_VARIABLE}': '{idempotency_key}'}}"

            async with session.get(
                f"{api_base_address}/v3/{domain}/events",
                params={
                    "recipient": recipient,
                    "user-variables": user_variables,
                    "event": "accepted",
                }
            ) as response:
                if response.status == 200:
                    body = await response.json()
                    if (len(body['items']) != 0):
                        logger.info(
                            "Found already accepted email message for "
                            f"recipient '{recipient}' with idempotency key "
                            f"'{idempotency_key}'"
                        )
                        # Since `items` is not empty, it means that the email
                        # with this idempotency key has already been sent.
                        return True
                else:
                    raise MailgunAPIError(
                        f"Something went wrong while running GET on Mailgun API (error code {response.status})"
                    )

        return False


# Mock version for testing with no actual email sending.
class MockMessageServicer(_AbstractMessageServicer):

    # The `emails_sent` list must be a class variable for tests to be able to
    # find and read it.
    emails_sent: ClassVar[dict[tuple[str, str], str]] = {}
    emails_sent_sema: ClassVar[asyncio.Semaphore] = asyncio.Semaphore(value=0)

    def __init__(self):
        super().__init__()
        # Invariant: there is only one `MockMessageServicer` per process at any
        # one time. Reset the class variable of sent emails to be empty when a
        # new instance is created, since it represents the start of a new
        # servicer for (likely) a new test.
        MockMessageServicer.emails_sent = {}
        MockMessageServicer.emails_sent_sema = asyncio.Semaphore(value=0)

    async def send_email(
        self,
        *,
        api_base_address: str,
        sender: str,
        recipient: str,
        subject: str,
        domain: str,
        idempotency_key: str,
        body: str,
        body_type: str,
    ):
        if (recipient, idempotency_key) not in MockMessageServicer.emails_sent:
            logger.info(
                f"Pretending to send email with subject {subject} to {recipient}"
            )
            MockMessageServicer.emails_sent[(recipient, idempotency_key)
                                           ] = body
            MockMessageServicer.emails_sent_sema.release()

    @classmethod
    async def is_email_with_idempotency_key_sent_to_recipient(
        self,
        *,
        api_base_address: str,
        recipient: str,
        domain: str,
        idempotency_key: str,
    ) -> bool:
        return list(MockMessageServicer.emails_sent.keys()).count(
            (recipient, idempotency_key)
        ) == 1

==> reboot_resemble-0.6.0-cp310-cp310-macosx_13_0_x86_64.whl: resemble/integrations/mailgun/servicers.py <==
import aiohttp
import asyncio
import os
import resemble.aio.auth.token_verifiers
import respect.logging
import uuid
from abc import abstractmethod
from datetime import datetime, timedelta
from google.protobuf.timestamp_pb2 import Timestamp
from resemble.aio.auth import Auth
from resemble.aio.auth.tokens import Token
from resemble.aio.contexts import ReaderContext, WriterContext
from resemble.aio.secrets import SecretNotFoundException, Secrets
from resemble.aio.tasks import TaskEffect
from resemble.integrations.mailgun import (
    MAILGUN_EVENT_API_CONSISTENCY_DELAY_SEC,
)
from resemble.integrations.mailgun.v1.mailgun_rsm import (
    Message,
    MessageState,
    SendRequest,
    SendResponse,
    SendTaskRequest,
    SendTaskResponse,
)
from resemble.v1alpha1.errors_pb2 import Ok, PermissionDenied, Unauthenticated
from typing import ClassVar, Optional

logger = respect.logging.get_logger(__name__)

IDEMPOTENCY_KEY_USER_VARIABLE = "idempotency_message_key"
MAILGUN_API_KEY_SECRET_NAME = "mailgun-api-key"
MAILGUN_API_BASE_ADDRESS_ENV_NAME = "MAILGUN_API_BASE_ADDRESS"

# If no API base address is injected for testing, use this one (the address for
# the real Mailgun API).
MAILGUN_API_BASE_ADDRESS = "https://api.mailgun.net"

# A "tag" used to ensure that the mailgun API key was correctly
# verified.
MAILGUN_API_KEY_VERIFIED = 'MAILGUN_API_KEY_VERIFIED'


class MailgunAPIError(RuntimeError):
    pass


class _AbstractMessageServicer(Message.Interface):

    class TokenVerifier(resemble.aio.auth.token_verifiers.TokenVerifier):

        def __init__(self, secrets: Optional[Secrets]):
            self._secrets = secrets

        async def verify_token(
            self,
            context: ReaderContext,
            token: str,
        ) -> Optional[Auth]:
            if self._secrets is None:
                # TODO(rjh): remove this case when secrets are also supported on
                #            the Cloud.
                return None
            try:
                api_key = await self._secrets.get(MAILGUN_API_KEY_SECRET_NAME)
            except SecretNotFoundException:
                logger.critical(
                    "Failed to verify token, missing secret "
                    f"'{MAILGUN_API_KEY_SECRET_NAME}'"
                )
                return None
            else:
                if token == api_key.decode():
                    return Auth(
                        # No user ID, the secret _is_ the mailgun API key.
                        user_id=None,
                        bearer_token=Token(
                            raw=token,
                            decoded={MAILGUN_API_KEY_VERIFIED: True},
                        ),
                    )
                else:
                    return None

    class Authorizer(Message.Authorizer):

        async def authorize(
            self,
            *,
            method_name: str,
            context: ReaderContext,
            state: Optional[Message.Authorizer.StateType] = None,
            request: Optional[Message.Authorizer.RequestTypes] = None,
        ) -> Message.Authorizer.Decision:
            if context.auth is None:
                return Unauthenticated()

            # Invariant: our `_AbstractMessageServicer.TokenVerifier`
            # is verifying that we have the correct API key, which is
            # sufficient for authorization, so all we need to do here
            # is ensure that we don't introduce a bug where a
            # different token verifier gets used. We do that by
            # looking for a specific field set by
            # `_AbstractMessageServicer.TokenVerifier`.
            assert context.auth.bearer_token is not None
            assert context.auth.bearer_token.decoded is not None
            if MAILGUN_API_KEY_VERIFIED in context.auth.bearer_token.decoded:
                return Ok()

            return PermissionDenied()

    def __init__(self):
        try:
            self._secrets: Optional[Secrets] = Secrets()
        except:
            # TODO(rjh,stuhood): remove this when secrets are supported on the
            #                    Cloud.
            # TODO(rjh,stuhood): make an error reading the Cloud API key more
            #                    specific than `Exception` so this use case
            #                    doesn't need to catch everything?
            logger.warning(
                "Failed to get secrets access, probably because secrets are "
                "not supported on the Reboot Cloud yet. Will not use secrets."
            )
            self._secrets = None

    def token_verifier(
        self,
    ) -> Optional[resemble.aio.auth.token_verifiers.TokenVerifier]:
        return _AbstractMessageServicer.TokenVerifier(self._secrets)

    def authorizer(self) -> Optional[Message.Authorizer]:
        return _AbstractMessageServicer.Authorizer()

    async def Send(
        self,
        context: WriterContext,
        request: SendRequest,
    ) -> Message.SendEffects:
        # Validate that we have exactly one of 'text' or 'html'
        # _before_ scheduling a task otherwise that task will never be
        # able to complete.
        body_type: Optional[str] = request.WhichOneof('body')

        if body_type is None:
            raise ValueError(
                "Exactly one of 'text' or 'html' should be specified"
            )

        initial_send_time = Timestamp()
        # GetCurrentTime() updates the Timestamp on which it is called in-place.
        initial_send_time.GetCurrentTime()

        # Generate an idempotency key to associate with the sending of this
        # message.
        state = MessageState(
            idempotency_key=str(uuid.uuid4()),
            initial_send_time=initial_send_time,
        )
        task: TaskEffect = self.schedule().SendTask(
            context,
            send_request=request,
        )
        return Message.SendEffects(
            state=state,
            tasks=[task],
            response=SendResponse(task_id=task.task_id),
        )

    async def SendTask(
        self,
        context: WriterContext,
        state: MessageState,
        request: SendTaskRequest,
    ) -> Message.SendTaskEffects:
        # Mailgun retains detailed event data for one day for free accounts and
        # up to 30 days for paid accounts based on subscription plan.
        # Since detailed event data is necessary for checking if an email has
        # already been sent, we cannot guarantee idempotency after more than a
        # day.
        #
        # Since we don't know if we're a free account or a paid
        # account, we conservatively bail out of this task if it's
        # been more than 24 hours since it first ran, to avoid
        # potentially sending this email twice.
        if (
            datetime.now()
            > state.initial_send_time.ToDatetime() + timedelta(hours=24)
        ):
            raise ValueError("Task is only valid for 24 hours after creation")

        body_type: Optional[str] = request.send_request.WhichOneof('body')

        assert body_type is not None

        body: Optional[str] = getattr(request.send_request, body_type)

        assert body is not None

        await self.send_email_idempotently(
            api_base_address=os.getenv(
                MAILGUN_API_BASE_ADDRESS_ENV_NAME, MAILGUN_API_BASE_ADDRESS
            ),
            sender=request.send_request.sender,
            recipient=request.send_request.recipient,
            subject=request.send_request.subject,
            domain=request.send_request.domain,
            idempotency_key=state.idempotency_key,
            body=body,
            body_type=body_type,
        )

        # Update accepted time to _now_, so when debugging we can see
        # when the message was accepted by mailgun (or if it has not
        # yet been accepted).
        #
        # `GetCurrentTime()` updates the `Timestamp` on which it is
        # called in-place.
        state.accepted_time.GetCurrentTime()

        return Message.SendTaskEffects(
            state=state,
            response=SendTaskResponse(),
        )

    @abstractmethod
    async def send_email(
        self,
        *,
        api_base_address: str,
        sender: str,
        recipient: str,
        subject: str,
        domain: str,
        idempotency_key: str,
        body: str,
        body_type: str,
    ):
        pass

    @abstractmethod
    async def is_email_with_idempotency_key_sent_to_recipient(
        self,
        *,
        api_base_address: str,
        recipient: str,
        domain: str,
        idempotency_key: str,
    ) -> bool:
        pass

    async def send_email_idempotently(
        self,
        *,
        api_base_address: str,
        sender: str,
        recipient: str,
        subject: str,
        domain: str,
        idempotency_key: str,
        body: str,
        body_type: str,
    ):
        # After our delay for eventual consistency (see comment on constant
        # declaration), we assume that the endpoint has fully caught up. If it
        # does not have evidence that the email has already been sent, we send
        # it now.
        await asyncio.sleep(MAILGUN_EVENT_API_CONSISTENCY_DELAY_SEC)

        if (
            not await self.is_email_with_idempotency_key_sent_to_recipient(
                api_base_address=api_base_address,
                recipient=recipient,
                domain=domain,
                idempotency_key=idempotency_key,
            )
        ):
            await self.send_email(
                api_base_address=api_base_address,
                sender=sender,
                recipient=recipient,
                subject=subject,
                domain=domain,
                body=body,
                body_type=body_type,
                idempotency_key=idempotency_key,
            )


class MessageServicer(_AbstractMessageServicer):

    def __init__(self):
        super().__init__()

    async def get_api_key(self) -> str:
        if self._secrets is None:
            # TODO(rjh): remove this when secrets are also supported on the cloud.
            return ""
        try:
            api_key = await self._secrets.get(MAILGUN_API_KEY_SECRET_NAME)
        except SecretNotFoundException as exception:
            logger.critical(f"No secret '{MAILGUN_API_KEY_SECRET_NAME}'")
            raise RuntimeError("No API key for mailgun") from exception
        return api_key.decode()

    async def send_email(
        self,
        *,
        api_base_address: str,
        sender: str,
        recipient: str,
        subject: str,
        domain: str,
        idempotency_key: str,
        body: str,
        body_type: str,
    ):
        """
        Sends an email with the given subject and body from the sender to
        the recipient, using the Mailgun API at the given api_base_address and
        domain. Authenticates to Mailgun with the given api_key credential.

        The email body can be provided either as plain text or as HTML content.
        We add the 'idempotency_key' as a user variable on the Mailgun message
        so that we can uniquely identify this email and ensure it is only sent
        once."""
        api_key = await self.get_api_key()

        async with aiohttp.ClientSession(
            auth=aiohttp.BasicAuth(login="api", password=api_key)
        ) as session:
            async with session.post(
                f"{api_base_address}/v3/{domain}/messages",
                data={
                    "from": sender,
                    "to": recipient,
                    "subject": subject,
                    body_type: body,
                    f"v:{IDEMPOTENCY_KEY_USER_VARIABLE}": idempotency_key,
                }
            ) as response:
                if response.status != 200:
                    raise MailgunAPIError(
                        f"Something went wrong while running POST on Mailgun API, error {response.status}"
                    )

                logger.info(
                    "Mailgun accepted email message for "
                    f"recipient '{recipient}' with idempotency key "
                    f"'{idempotency_key}'"
                )

    async def is_email_with_idempotency_key_sent_to_recipient(
        self,
        *,
        api_base_address: str,
        recipient: str,
        domain: str,
        idempotency_key: str,
    ) -> bool:
        api_key = await self.get_api_key()

        async with aiohttp.ClientSession(
            auth=aiohttp.BasicAuth(login="api", password=api_key)
        ) as session:
            user_variables = f"{{'{IDEMPOTENCY_KEY_USER_VARIABLE}': '{idempotency_key}'}}"

            async with session.get(
                f"{api_base_address}/v3/{domain}/events",
                params={
                    "recipient": recipient,
                    "user-variables": user_variables,
                    "event": "accepted",
                }
            ) as response:
                if response.status == 200:
                    body = await response.json()
                    if (len(body['items']) != 0):
                        logger.info(
                            "Found already accepted email message for "
                            f"recipient '{recipient}' with idempotency key "
                            f"'{idempotency_key}'"
                        )
                        # Since `items` is not empty, it means that the email
                        # with this idempotency key has already been sent.
                        return True
                else:
                    raise MailgunAPIError(
                        f"Something went wrong while running GET on Mailgun API (error code {response.status})"
                    )

        return False


# Mock version for testing with no actual email sending.
class MockMessageServicer(_AbstractMessageServicer):

    # The `emails_sent` list must be a class variable for tests to be able to
    # find and read it.
    emails_sent: ClassVar[list[dict[str, str]]] = []
    emails_sent_sema: ClassVar[asyncio.Semaphore] = asyncio.Semaphore(value=0)
    # Note: we track idempotency of sent email independently from message bodies to avoid
    # callers needing to worry about races between clearing the `emails_sent` list, and
    # tasks (re)running in the background.
    _idempotency_keys: ClassVar[set[tuple[str, str]]] = set()

    def __init__(self):
        super().__init__()
        # Invariant: there is only one `MockMessageServicer` per process at any
        # one time. Reset the class variable of sent emails to be empty when a
        # new instance is created, since it represents the start of a new
        # servicer for (likely) a new test.
        MockMessageServicer.emails_sent = []
        MockMessageServicer.emails_sent_sema = asyncio.Semaphore(value=0)
        MockMessageServicer._idempotency_keys = set()

    async def send_email(
        self,
        *,
        recipient: str,
        subject: str,
        idempotency_key: str,
        **kwargs: str,
    ):
        key = (recipient, idempotency_key)
        if key in MockMessageServicer._idempotency_keys:
            return
        MockMessageServicer._idempotency_keys.add(key)

        logger.info(
            f"Pretending to send email with subject {subject} to {recipient}"
        )
        MockMessageServicer.emails_sent.append(
            dict(**kwargs, recipient=recipient, subject=subject)
        )
        MockMessageServicer.emails_sent_sema.release()

    @classmethod
    async def is_email_with_idempotency_key_sent_to_recipient(
        self,
        *,
        api_base_address: str,
        recipient: str,
        domain: str,
        idempotency_key: str,
    ) -> bool:
        return (
            recipient, idempotency_key
        ) in MockMessageServicer._idempotency_keys
